/*
 * b2c2-usb-core.c - USB access methods and functions for
 *                   B2C2/Technisat USB DVB devices
 *
 * Copyright (C) 2004 Patrick Boettcher <patrick.boettcher@desy.de>,
 *                    Bjarne Steinsbo <bjarne@steinsbo.com> and
 *                    Luca Bertagnolio
 *
 * based on information provided by John Jurrius from BBTI, Inc.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#include <linux/usb.h>
#include "flexcop-common.h"

static int debug;
module_param(debug, int, 0644);
MODULE_PARM_DESC(debug, "set debugging level (1=info,ts=2,ctrl=4 (or-able)).");
static int enable_hw_filters = 2;
module_param(enable_hw_filters, int, 0444);
MODULE_PARM_DESC(enable_hw_filters, "enable hardware filters: supported values: 0 (none), 1, 2");

/* Version information */
#define DRIVER_VERSION "0.0"
#define DRIVER_DESC "Driver for B2C2/Technisat Air/Cable/Sky-2-PC USB devices"
#define DRIVER_AUTHOR "Patrick Boettcher, patrick.boettcher@desy.de"

/* transfer parameters */
#define B2C2_USB_FRAMES_PER_ISO		4
#define B2C2_USB_NUM_ISO_URB		4    /* TODO check out a good value */

#define B2C2_USB_CTRL_PIPE_IN		usb_rcvctrlpipe(b2c2->udev,0)
#define B2C2_USB_CTRL_PIPE_OUT		usb_sndctrlpipe(b2c2->udev,0)
#define B2C2_USB_DATA_PIPE			usb_rcvisocpipe(b2c2->udev,0x81)

/* debug */
#define dprintk(level,args...) \
	    do { if ((debug & level)) { printk(args); } } while (0)
#define debug_dump(b,l) if (debug) {\
	int i; deb_xfer("%s: %d > ",__FUNCTION__,l); \
	for (i = 0; i < l; i++) deb_xfer("%02x ", b[i]); \
	deb_xfer("\n");\
}

#define deb_info(args...) dprintk(0x01,args)
#define deb_ts(args...)   dprintk(0x02,args)
#define deb_ctrl(args...) dprintk(0x04,args)
#define deb_i2c(args...) dprintk(0x08,args)

struct usb_b2c2_usb {
	struct usb_device *udev;
	struct usb_interface *uintf;

	u8 *iso_buffer;
	int buffer_size;
	dma_addr_t iso_dma_handle;
	struct urb *iso_urb[B2C2_USB_NUM_ISO_URB];
};


/*
 * USB
 * 10 90 34 12 78 56 04 00
 * usb_control_msg(udev, usb_sndctrlpipe(udev,0),
 * 0x90,
 * 0x10,
 * 0x1234,
 * 0x5678,
 * buf,
 * 4,
 * 5*HZ);
 *
 * extern int usb_control_msg(struct usb_device *dev, unsigned int pipe,
 * __u8 request,
 * __u8 requesttype,
 * __u16 value,
 * __u16 index,
 * void *data,
 * __u16 size,
 * int timeout);
 *
 */

/* request types TODO What is its use?*/
typedef enum {

/* something is wrong with this part
	RTYPE_READ_DW         = (1 << 6),
	RTYPE_WRITE_DW_1      = (3 << 6),
	RTYPE_READ_V8_MEMORY  = (6 << 6),
	RTYPE_WRITE_V8_MEMORY = (7 << 6),
	RTYPE_WRITE_V8_FLASH  = (8 << 6),
	RTYPE_GENERIC         = (9 << 6),
*/
	RTYPE_READ_DW = (3 << 6),
	RTYPE_WRITE_DW_1 = (1 << 6),

	RTYPE_READ_V8_MEMORY  = (6 << 6),
	RTYPE_WRITE_V8_MEMORY = (7 << 6),
	RTYPE_WRITE_V8_FLASH  = (8 << 6),
	RTYPE_GENERIC         = (9 << 6),
} b2c2_usb_request_type_t;

/* request */
typedef enum {
	B2C2_USB_WRITE_V8_MEM = 0x04,
	B2C2_USB_READ_V8_MEM  = 0x05,
	B2C2_USB_READ_REG     = 0x08,
	B2C2_USB_WRITE_REG    = 0x0A,
/*	B2C2_USB_WRITEREGLO   = 0x0A, */
	B2C2_USB_WRITEREGHI   = 0x0B,
	B2C2_USB_FLASH_BLOCK  = 0x10,
	B2C2_USB_I2C_REQUEST  = 0x11,
	B2C2_USB_UTILITY      = 0x12,
} b2c2_usb_request_t;

/* function definition for I2C_REQUEST */
typedef enum {
	USB_FUNC_I2C_WRITE       = 0x01,
	USB_FUNC_I2C_MULTIWRITE  = 0x02,
	USB_FUNC_I2C_READ        = 0x03,
	USB_FUNC_I2C_REPEATWRITE = 0x04,
	USB_FUNC_GET_DESCRIPTOR  = 0x05,
	USB_FUNC_I2C_REPEATREAD  = 0x06,
/* DKT 020208 - add this to support special case of DiSEqC */
	USB_FUNC_I2C_CHECKWRITE  = 0x07,
	USB_FUNC_I2C_CHECKRESULT = 0x08,
} b2c2_usb_i2c_function_t;

/*
 * function definition for UTILITY request 0x12
 * DKT 020304 - new utility function
 */
typedef enum {
	UTILITY_SET_FILTER          = 0x01,
	UTILITY_DATA_ENABLE         = 0x02,
	UTILITY_FLEX_MULTIWRITE     = 0x03,
	UTILITY_SET_BUFFER_SIZE     = 0x04,
	UTILITY_FLEX_OPERATOR       = 0x05,
	UTILITY_FLEX_RESET300_START = 0x06,
	UTILITY_FLEX_RESET300_STOP  = 0x07,
	UTILITY_FLEX_RESET300       = 0x08,
	UTILITY_SET_ISO_SIZE        = 0x09,
	UTILITY_DATA_RESET          = 0x0A,
	UTILITY_GET_DATA_STATUS     = 0x10,
	UTILITY_GET_V8_REG          = 0x11,
/* DKT 020326 - add function for v1.14 */
	UTILITY_SRAM_WRITE          = 0x12,
	UTILITY_SRAM_READ           = 0x13,
	UTILITY_SRAM_TESTFILL       = 0x14,
	UTILITY_SRAM_TESTSET        = 0x15,
	UTILITY_SRAM_TESTVERIFY     = 0x16,
} b2c2_usb_utility_function_t;

#define B2C2_WAIT_FOR_OPERATION_RW  1  /* 1 s */
#define B2C2_WAIT_FOR_OPERATION_RDW 3  /* 3 s */
#define B2C2_WAIT_FOR_OPERATION_WDW 1  /* 1 s */

#define B2C2_WAIT_FOR_OPERATION_V8READ   3  /* 3 s */
#define B2C2_WAIT_FOR_OPERATION_V8WRITE  3  /* 3 s */
#define B2C2_WAIT_FOR_OPERATION_V8FLASH  3  /* 3 s */

/* JLP 111700: we will include the 1 bit gap between the upper and lower 3 bits
 * in the IBI address, to make the V8 code simpler.
 * PCI ADDRESS FORMAT: 0x71C -> 0000 0111 0001 1100 (these are the six bits used)
 *                  in general: 0000 0HHH 000L LL00
 * IBI ADDRESS FORMAT:                    RHHH BLLL
 *
 * where R is the read(1)/write(0) bit, B is the busy bit
 * and HHH and LLL are the two sets of three bits from the PCI address.
 */
#define B2C2_FLEX_PCIOFFSET_TO_INTERNALADDR(usPCI) (u8) (((usPCI >> 2) & 0x07) + ((usPCI >> 4) & 0x70))
#define B2C2_FLEX_INTERNALADDR_TO_PCIOFFSET(ucAddr) (u16) (((ucAddr & 0x07) << 2) + ((ucAddr & 0x70) << 4))

/*
 * DKT 020228 
 * - forget about this VENDOR_BUFFER_SIZE, read and write register
 *   deal with DWORD or 4 bytes, that should be should from now on
 * - from now on, we don't support anything older than firm 1.00
 *   I eliminated the write register as a 2 trip of writing hi word and lo word
 *   and force this to write only 4 bytes at a time.
 *   NOTE: this should work with all the firmware from 1.00 and newer
 */
static int b2c2_usb_readwrite_dw(struct b2c2_device *card, u16 wRegOffsPCI, u32 *val, u8 read)
{
	struct usb_b2c2_usb *b2c2 = card->bus_specific;
	u8 request = read ? B2C2_USB_READ_REG : B2C2_USB_WRITE_REG;
	u8 request_type = (read ? USB_DIR_IN : USB_DIR_OUT) | USB_TYPE_VENDOR;
	u8 wAddress = B2C2_FLEX_PCIOFFSET_TO_INTERNALADDR(wRegOffsPCI) | (read ? 0x80 : 0);

	int len = usb_control_msg(b2c2->udev,
			read ? B2C2_USB_CTRL_PIPE_IN : B2C2_USB_CTRL_PIPE_OUT,
			request,
			request_type,  /* 0xc0 read or 0x40 write*/
			wAddress,
			0,
			val,
			sizeof(u32),
			B2C2_WAIT_FOR_OPERATION_RDW * HZ);

	if (len != sizeof(u32)) {
		b2c2_err("error while %s dword from %d (%d).",read ? "reading" : "writing",
			wAddress,wRegOffsPCI);
		return -EIO;
	} 
	return 0;
}

/*
 * DKT 010817 - add support for V8 memory read/write and flash update
 */
static int b2c2_usb_v8_memory_req(struct usb_b2c2_usb *b2c2,
		b2c2_usb_request_t req, u8 page, u16 wAddress,
		u16 buflen, u8 *pbBuffer)
{
//	u8 dwRequestType;
	u8 request_type = USB_TYPE_VENDOR;
	u16 wIndex;
	int nWaitTime,pipe,len;

	wIndex = page << 8;

	switch (req) {
		case B2C2_USB_READ_V8_MEM:
			nWaitTime = B2C2_WAIT_FOR_OPERATION_V8READ;
			request_type |= USB_DIR_IN;
//			dwRequestType = (u8) RTYPE_READ_V8_MEMORY;
			pipe = B2C2_USB_CTRL_PIPE_IN;
		break;
		case B2C2_USB_WRITE_V8_MEM:
			wIndex |= pbBuffer[0];
			request_type |= USB_DIR_OUT;
			nWaitTime = B2C2_WAIT_FOR_OPERATION_V8WRITE;
//			dwRequestType = (u8) RTYPE_WRITE_V8_MEMORY;
			pipe = B2C2_USB_CTRL_PIPE_OUT;
		break;
		case B2C2_USB_FLASH_BLOCK:
			request_type |= USB_DIR_OUT;
			nWaitTime = B2C2_WAIT_FOR_OPERATION_V8FLASH;
//			dwRequestType = (u8) RTYPE_WRITE_V8_FLASH;
			pipe = B2C2_USB_CTRL_PIPE_OUT;
		break;
		default:
			deb_info("unsupported request for v8_mem_req %x.\n",req);
		return -EINVAL;
	}
	len = usb_control_msg(b2c2->udev,pipe,
			req,
			request_type,
			wAddress,
			wIndex,
			pbBuffer,
			buflen,
			nWaitTime * HZ);
	return len == buflen ? 0 : -EIO;
}

#define V8_MEMORY_PAGE_SIZE     0x8000      // 32K
#define V8_MEMORY_PAGE_MASK     0x7FFF
#define V8_MEMORY_PAGE_SHIFT    16

#define V8_MEMORY_EXT_ENABLE    0x00008000
#define V8_MEMORY_SYS_RAM       0x00000000
#define V8_MEMORY_EXT_RAM       0x00008000
#define V8_MEMORY_DVB_CI        0x00208000
#define V8_MEMORY_DVB_DS        0x00408000
#define V8_MEMORY_MULTI2        0x00608000
#define V8_MEMORY_FLASH         0x00808000
#define V8_MEMORY_FLASH_WRITE   0x00000000

#define USB_MEM_WRITE_MAX       1
#define USB_MEM_READ_MAX        32
#define USB_FLASH_MAX           8

#define v8Address(dwPageStart,dwAddress) ( \
		dwPageStart + \
		((dwAddress / V8_MEMORY_PAGE_SIZE) << V8_MEMORY_PAGE_SHIFT) + \
		 (dwAddress & V8_MEMORY_PAGE_MASK) )

#define v8FlatAddr(dwAddress) ( \
		((dwAddress >> V8_MEMORY_PAGE_SHIFT) * V8_MEMORY_PAGE_SIZE) + \
		 (dwAddress & V8_MEMORY_PAGE_MASK) )


static int b2c2_usb_memory_raw_req(struct b2c2_device *b2c2,u8 req,u32 dwAddr,u8 *buf,u16 len)
{
	return b2c2_usb_v8_memory_req(b2c2->bus_specific,req,
			(u8) (dwAddr >> 16), (u16) (dwAddr & 0xFFFF), 
			len, buf);
}

static int b2c2_usb_memory_req(struct b2c2_device *b2c2,u8 req,u8 dwAddress,void *pvData, u32 len)
{
	int ret = 0;
	u16 wMax;
	u32 dwExtEnable = dwAddress & V8_MEMORY_EXT_ENABLE;
	u8 *data = (u8 *) pvData;
	u32 dwFlatAddr = v8FlatAddr(dwAddress);
	u32 dwPageChunk, dwChunk;

	switch (req) {
		case B2C2_USB_READ_V8_MEM:  wMax = USB_MEM_READ_MAX; break;
		case B2C2_USB_WRITE_V8_MEM: wMax = USB_MEM_WRITE_MAX; break;
		case B2C2_USB_FLASH_BLOCK:  wMax = USB_FLASH_MAX; break;
		default: ret = -EINVAL;
	}
	
	for (; ret == 0 && len; dwFlatAddr += dwPageChunk) {
		u16 wCount;
		u32 dwPage1, dwPage2;

		dwPage1 = dwFlatAddr / V8_MEMORY_PAGE_SIZE;
		dwPage2 = ( dwFlatAddr + len - 1 ) / V8_MEMORY_PAGE_SIZE;
		if ( dwPage1 == dwPage2 ) 
			dwChunk = dwPageChunk = len;
		else
			dwChunk = dwPageChunk = ( dwPage1 + 1 ) * V8_MEMORY_PAGE_SIZE - dwFlatAddr;
		
		len -= dwPageChunk;
		dwAddress = (dwPage1 << V8_MEMORY_PAGE_SHIFT) | dwExtEnable | (dwFlatAddr & V8_MEMORY_PAGE_MASK);
		for ( ;ret == 0 && dwPageChunk; dwAddress+=wCount, dwPageChunk-=wCount, data+=wCount) {
			wCount = dwPageChunk < wMax ? dwPageChunk : wMax;
			ret = b2c2_usb_memory_raw_req(b2c2,req,dwAddress, data, wCount);
		}
		dwFlatAddr += dwChunk;
	}
	return ret;
}

/* useful shortcuts */
#define b2c2_usb_memory_writeb(b2c2,dwAddress,bData) \
	b2c2_usb_memory_req(b2c2,B2C2_USB_WRITE_V8_MEM,dwAddress,&bData,1)

u8 b2c2_usb_memory_readb(struct b2c2_device *b2c2, u32 dwAddress)
{
	u8 b;
	b2c2_usb_memory_req(b2c2,B2C2_USB_READ_V8_MEM,dwAddress,&b,1);
	return b;
}

#define b2c2_usb_flash_readb(b2c2,dwAddress) \
	b2c2_usb_memory_readb(b2c2, v8Address(V8_MEMORY_FLASH,dwAddress))

/*
 *
 * TODO
 * BYTE DrvFlashRead( DWORD dwAddress,void *pvData,DWORD dwLength )
 * {
 *     return UsbDrvMemoryRead( v8Address( V8_MEMORY_FLASH,dwAddress ),pvData,dwLength );
 *     }
 *
 *     BYTE DrvFlashReadB( DWORD dwAddress )
 *     {
 *         return UsbDrvMemoryReadB( v8Address( V8_MEMORY_FLASH,dwAddress ) );
 *         }
 *
 *         BYTE DrvFlashWrite( DWORD dwAddress,void *pvData,DWORD dwLength )
 *         {
 *             return UsbDrvFlashWrite( v8Address( V8_MEMORY_FLASH_WRITE,dwAddress ),
 *                                          (BYTE *)pvData,dwLength );
 *                                          }
 */


static int b2c2_usb_utility_req(struct usb_b2c2_usb *b2c2, int set,
		b2c2_usb_utility_function_t func, u8 extra, u16 wIndex,
		u16 buflen, u8 *pvBuffer)
{
	u16 wValue;
	u8 request_type = (set ? USB_DIR_OUT : USB_DIR_IN) | USB_TYPE_VENDOR;
//	u8 dwRequestType = (u8) RTYPE_GENERIC,
	int nWaitTime = 2,
		pipe = set ? B2C2_USB_CTRL_PIPE_OUT : B2C2_USB_CTRL_PIPE_IN,
		len;

	wValue = (func << 8) | extra;

	len = usb_control_msg(b2c2->udev,pipe,
			B2C2_USB_UTILITY,
			request_type,			
			wValue,
			wIndex,
			pvBuffer,
			buflen,
			nWaitTime * HZ);
	return len == buflen ? 0 : -EIO;
}

/* usb i2c stuff */
static int b2c2_usb_i2c_req(struct usb_b2c2_usb *b2c2,
		b2c2_usb_request_t req, b2c2_usb_i2c_function_t func,
		u8 port, u8 chipaddr, u8 addr, u8 buflen, u8 *buf)
{
	u16 wValue, wIndex;
	int nWaitTime,pipe,len;
//	u8 dwRequestType;
	u8 request_type = USB_TYPE_VENDOR;

	switch (func) {
		case USB_FUNC_I2C_WRITE:
		case USB_FUNC_I2C_MULTIWRITE:
		case USB_FUNC_I2C_REPEATWRITE:
		/* DKT 020208 - add this to support special case of DiSEqC */
		case USB_FUNC_I2C_CHECKWRITE:
			pipe = B2C2_USB_CTRL_PIPE_OUT;
			nWaitTime = 2;
//			dwRequestType = (u8) RTYPE_GENERIC;
			request_type |= USB_DIR_OUT;
		break;
		case USB_FUNC_I2C_READ:
		case USB_FUNC_I2C_REPEATREAD:
			pipe = B2C2_USB_CTRL_PIPE_IN;
			nWaitTime = 2;
//			dwRequestType = (u8) RTYPE_GENERIC;
			request_type |= USB_DIR_IN;
		break;
		default:
			deb_info("unsupported function for i2c_req %x\n",func);
			return -EINVAL;
	}
	wValue = (func << 8 ) | port;
	wIndex = (chipaddr << 8 ) | addr;

	len = usb_control_msg(b2c2->udev,pipe,
			req,
			request_type,
			wValue,
			wIndex,
			buf,
			buflen,
			nWaitTime * HZ);

	return len == buflen ? 0 : -EREMOTEIO;
}

static u32 b2c2_usb_read_dw_old(struct b2c2_device *card, u16 wRegOffsPCI)
{
	err("do not use this functions. ");
	return -EINVAL;
}

static int b2c2_usb_write_dw_old(struct b2c2_device *card, u16 wRegOffsPCI, u32 val)
{
	err("do not use this functions. ");
	return -EINVAL;
}

/* actual bus specific access functions, make sure prototype are/will be equal to pci */
static flexcop_ibi_value b2c2_usb_read_ibi_reg(struct b2c2_device *card, flexcop_ibi_register reg)
{
	flexcop_ibi_value val;
	val.raw = 0;
	b2c2_usb_readwrite_dw(card,reg, &val.raw, 1);
	return val;
}

static int b2c2_usb_write_ibi_reg(struct b2c2_device *card, flexcop_ibi_register reg, flexcop_ibi_value val)
{
	return b2c2_usb_readwrite_dw(card,reg, &val.raw, 0);
}

static int b2c2_usb_i2c_read(struct b2c2_device *card, u8 port, u8 chipaddr, u8 addr, u8 *buf, u16 len)
{
	return b2c2_usb_i2c_req(card->bus_specific,B2C2_USB_I2C_REQUEST,
			USB_FUNC_I2C_READ, port, chipaddr, addr, len, buf) == 0 ? len : 0;
}

static int b2c2_usb_i2c_write(struct b2c2_device *card, u8 port, u8 chipaddr, u8 addr, u8 *buf, u16 len)
{
	return b2c2_usb_i2c_req(card->bus_specific,B2C2_USB_I2C_REQUEST, 
			USB_FUNC_I2C_WRITE, port, chipaddr, addr, len, buf) == 0 ? len : 0;
}
		
static void b2c2_dumpfourreg(struct b2c2_device *card, u16 offs)
{
	flexcop_ibi_value r0,r1,r2,r3;
	r0 = b2c2_usb_read_ibi_reg(card, offs);
	r1 = b2c2_usb_read_ibi_reg(card, offs + 0x04);
	r2 = b2c2_usb_read_ibi_reg(card, offs + 0x08);
	r3 = b2c2_usb_read_ibi_reg(card, offs + 0x0c);
	deb_ctrl("dump: offset: %03x, %08x, %08x, %08x, %08x\n",offs,r0.raw,r1.raw,r2.raw,r3.raw);
}

static void b2c2_urb_complete(struct urb *urb, struct pt_regs *ptregs)
{
	struct usb_b2c2_usb *b2c2 = urb->context;
	deb_ts("urb completed, bufsize: %d actlen; %d\n",urb->transfer_buffer_length, urb->actual_length);

//	usb_submit_urb(urb,GFP_ATOMIC); // enable for real action
}

static void open_stream_usb(struct b2c2_device *card, u16 pid)
{
	/* TODO */
}

static void close_stream_usb(struct b2c2_device *card, u16 pid)
{
	/* TODO */
}

static void b2c2_exit_usb(struct usb_b2c2_usb *b2c2)
{
	int i;
	for (i = 0; i < B2C2_USB_NUM_ISO_URB; i++)
		if (b2c2->iso_urb[i] != NULL) { /* not sure about unlink_urb and iso-urbs TODO */
			deb_info("unlinking/killing urb no. %d\n",i);
			usb_kill_urb(b2c2->iso_urb[i]);
			usb_free_urb(b2c2->iso_urb[i]);
		}

	if (b2c2->iso_buffer != NULL)
		pci_free_consistent(NULL,b2c2->buffer_size, b2c2->iso_buffer, b2c2->iso_dma_handle);

}

static int b2c2_init_usb(struct usb_b2c2_usb *b2c2)
{
	u16 frame_size = b2c2->uintf->cur_altsetting->endpoint[0].desc.wMaxPacketSize;
	int bufsize = B2C2_USB_NUM_ISO_URB * B2C2_USB_FRAMES_PER_ISO * frame_size,i,j,ret;
	int buffer_offset = 0;

	deb_info("creating %d iso-urbs with %d frames each of %d bytes size = %d.\n",
			B2C2_USB_NUM_ISO_URB, B2C2_USB_FRAMES_PER_ISO, frame_size,bufsize);

	b2c2->iso_buffer = pci_alloc_consistent(NULL,bufsize,&b2c2->iso_dma_handle);
	if (b2c2->iso_buffer == NULL)
		return -ENOMEM;
	memset(b2c2->iso_buffer, 0, bufsize);
	b2c2->buffer_size = bufsize;

	/* creating iso urbs */
	for (i = 0; i < B2C2_USB_NUM_ISO_URB; i++)
		if (!(b2c2->iso_urb[i] = usb_alloc_urb(B2C2_USB_FRAMES_PER_ISO,GFP_ATOMIC))) {
			ret = -ENOMEM;
			goto urb_error;
		}
	/* initialising and submitting iso urbs */
	for (i = 0; i < B2C2_USB_NUM_ISO_URB; i++) {
		int frame_offset = 0;
		struct urb *urb = b2c2->iso_urb[i];
		deb_info("initializing and submitting urb no. %d (buf_offset: %d).\n",i,buffer_offset);

		urb->dev = b2c2->udev;
		urb->context = b2c2;
		urb->complete = b2c2_urb_complete;
		urb->pipe = B2C2_USB_DATA_PIPE;
		urb->transfer_flags = URB_ISO_ASAP;
		urb->interval = 1;
		urb->number_of_packets = B2C2_USB_FRAMES_PER_ISO;
		urb->transfer_buffer_length = frame_size * B2C2_USB_FRAMES_PER_ISO;
		urb->transfer_buffer = b2c2->iso_buffer + buffer_offset;

		buffer_offset += frame_size * B2C2_USB_FRAMES_PER_ISO;
		for (j = 0; j < B2C2_USB_FRAMES_PER_ISO; j++) {
			deb_info("urb no: %d, frame: %d, frame_offset: %d\n",i,j,frame_offset);
			urb->iso_frame_desc[j].offset = frame_offset;
			urb->iso_frame_desc[j].length = frame_size;
			frame_offset += frame_size;
		}

		if ((ret = usb_submit_urb(b2c2->iso_urb[i],GFP_ATOMIC))) {
			b2c2_err("submitting urb %d failed with %d.",i,ret);
			goto urb_error;
		}
		deb_info("submitted urb no. %d.\n",i);
	}

	ret = 0;
	goto success;
urb_error:
	b2c2_exit_usb(b2c2);
success:
	return ret;
}

static int b2c2_usb_probe(struct usb_interface *intf,
		const struct usb_device_id *id)
{
	struct usb_device *udev = interface_to_usbdev(intf);
	struct usb_b2c2_usb *b2c2 = NULL;
	struct b2c2_device *card = NULL;
	int ret;

	switch (udev->speed) {
		case USB_SPEED_LOW:
			b2c2_err("cannot handle USB speed because it is to sLOW.");
			break;
		case USB_SPEED_FULL:
			b2c2_info("running at FULL speed.");
			break;
		case USB_SPEED_HIGH:
			b2c2_info("running at HIGH speed.");
			break;
		case USB_SPEED_UNKNOWN: /* fall through */
		default:
			b2c2_err("cannot handle USB speed because it is unkown.");
			ret = -EPROTO;
			goto usb_init_error;
	}

	card = b2c2_device_kmalloc(sizeof(struct usb_b2c2_usb));
	if (card == NULL) {
		return -ENOMEM;
	}

	b2c2 = card->bus_specific;

	b2c2->udev = udev;
	b2c2->uintf = intf;

	card->read_dw_old   = b2c2_usb_read_dw_old;
	card->write_dw_old  = b2c2_usb_write_dw_old;
	
	card->read_ibi_reg  = b2c2_usb_read_ibi_reg;
	card->write_ibi_reg = b2c2_usb_write_ibi_reg;
	
	card->i2c_read  = b2c2_usb_i2c_read;
	card->i2c_write = b2c2_usb_i2c_write;
	card->open_stream = open_stream_usb;
	card->close_stream = close_stream_usb;

	/* Copy some info data from device specific to device independant */
	card->debug = debug;
	card->enable_hw_filters = enable_hw_filters;
	card->dev_vendor = udev->descriptor.idVendor;
	card->dev_device = udev->descriptor.idProduct;
	card->dev_subsystem_vendor = 0;
	card->dev_subsystem_device = 0;

	/* use the alternate setting with the larges buffer */
	usb_set_interface(udev,0,1);

	if ((ret = b2c2_init_usb(b2c2)))
		goto usb_init_error;

	usb_set_intfdata(intf, card);

	b2c2_dumpfourreg(card, 0x200);
	b2c2_dumpfourreg(card, 0x300);
	b2c2_dumpfourreg(card, 0x400);
	b2c2_dumpfourreg(card, 0x700);

	b2c2_flexcop_reset(card);
	ret = b2c2_initialize(card);

	if (ret != 0) {
		b2c2_info("%s error while loading driver (%d)",DRIVER_DESC,ret);
		goto usb_init_error;
	}

	b2c2_info("%s successfully initialized and connected.",DRIVER_DESC);
	ret = 0;
	goto success;

usb_init_error:
	kfree(b2c2);
	kfree(card);
success:
	return ret;
}

static void b2c2_usb_disconnect(struct usb_interface *intf)
{
	struct b2c2_device *card = usb_get_intfdata(intf);
	usb_set_intfdata(intf,NULL);
	if (card != NULL) {
		b2c2_uninitialize(card);
		b2c2_exit_usb(card->bus_specific);
		kfree(card->bus_specific);
		kfree(card);
	}
	b2c2_info("%s successfully deinitialized and disconnected.",DRIVER_DESC);
}

static struct usb_device_id b2c2_usb_table [] = {
	    { USB_DEVICE(0x0af7, 0x0101) }
};

/* usb specific object needed to register this driver with the usb subsystem */
static struct usb_driver b2c2_usb_driver = {
	.owner		= THIS_MODULE,
	.name		= "dvb_b2c2_usb",
	.probe 		= b2c2_usb_probe,
	.disconnect = b2c2_usb_disconnect,
	.id_table 	= b2c2_usb_table,
};

/* module stuff */
static int __init b2c2_usb_init(void)
{
	int result;
	if ((result = usb_register(&b2c2_usb_driver))) {
		b2c2_err("usb_register failed. Error number %d",result);
		return result;
	}

	return 0;
}

static void __exit b2c2_usb_exit(void)
{
	/* deregister this driver from the USB subsystem */
	usb_deregister(&b2c2_usb_driver);
}

module_init(b2c2_usb_init);
module_exit(b2c2_usb_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
